package org.openxma.dsl.generator.component;

import static org.openxma.dsl.generator.GeneratorProperties.getModelModifier;

import java.util.Collection;

import org.eclipse.emf.mwe.core.WorkflowContext;
import org.eclipse.emf.mwe.core.issues.Issues;
import org.eclipse.emf.mwe.core.lib.WorkflowComponentWithModelSlot;
import org.eclipse.emf.mwe.core.monitor.ProgressMonitor;
import org.openxma.dsl.core.model.Model;

/**
 * Workflow component used for pre-processing the contents of the configured inputslot models.
 */
public class DomainModelModifierComponent extends WorkflowComponentWithModelSlot {
	private Object modelObject;

	public DomainModelModifierComponent() {
		super();
		setSkipOnErrors(true);
	}

	public void invokeInternal(final WorkflowContext ctx, final ProgressMonitor monitor, final Issues issues) {
		modelObject = ctx.get(getModelSlot());
		if (modelObject == null) {
			issues.addWarning(this, "content of modelSlot " + getModelSlot() + " is null.");
		}
		doModification(ctx, monitor, issues, modelObject);
	}

	@SuppressWarnings("unchecked")
	protected void doModification(WorkflowContext ctx, ProgressMonitor monitor, Issues issues, Object modelSlot) {
		if (modelSlot instanceof Collection) {
			Collection<Model> models = (Collection<Model>) modelSlot;
			for (Model model : models) {
				getModelModifier().modifyModel((Model) model);
			}
		}
	}
	
	@Override
	public String getLogMessage() {
		return "Run model modifier '"+getModelModifier().getClass()+"'";
	}
}
