package org.openxma.dsl.generator.component;

import java.io.File;

import org.eclipse.xpand2.output.FileHandle;
import org.eclipse.xpand2.output.PostProcessor;
import org.openxma.dsl.common.logging.LogUtil;
import org.openxma.dsl.generator.GeneratorProperties;
import org.openxma.hbmfilemerge.HbmFileMerge;
import org.openxma.hbmfilemerge.HbmFileMergeInput;

public class HbmFileMergePostProcessor implements PostProcessor {
    private File outletResources, outletGeneratedResources;

    @SuppressWarnings("deprecation")
    public void afterClose(FileHandle impl) {
        initPath();
        File genHbm = impl.getTargetFile();

        // because the gen file is created after the normal hbm file we react to
        // the creation of gen hbm files
        // because then we can merge into the normal hbm files without loosing
        // the merge result
        if (!isExitistingHbmGenFileInGenResourcesFolder(genHbm))
            return;
        try {
            HbmFileMerge.getInstance(outletResources, outletGeneratedResources).mergeFragmentHbmWith(genHbm);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private void initPath() {
        if (null != outletResources && null != outletGeneratedResources)
            return;

        outletResources = new File(GeneratorProperties.getProperty(GeneratorProperties.OUTLET_RESOURCES));
        outletGeneratedResources = new File(GeneratorProperties
                .getProperty(GeneratorProperties.OUTLET_GENERATED_RESOURCES));
    }

    /**
     * @param delta
     *            , the delta of the changed resource
     * @return <code>true</code> if the changed recource/file has the file name
     *         suffix
     *         <code>HbmFileChangeListener.FRAGMENT_HBM_FILE_SUFFIX</code> and
     *         is within the Java Source folder
     */
    private boolean isExitistingHbmGenFileInGenResourcesFolder(File toCheck) {
        if (!toCheck.exists() || !toCheck.isFile()) {
            LogUtil.info("file does not exist or is not a file: \n" + toCheck.getAbsolutePath());
            return false;
        }

        return (HbmFileMergeInput.isGenHbmFile(toCheck.getName()) && isWithinGenResourceFolder(toCheck)) ? true : false;
    }

    private boolean isWithinGenResourceFolder(File toCheck) {
        return toCheck.getAbsolutePath().startsWith(outletGeneratedResources.getAbsolutePath()) ? true : false;
    }

    public void beforeWriteAndClose(FileHandle impl) {
    }

}
