package org.openxma.dsl.generator.helper;

import java.util.List;
import java.util.Stack;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.openxma.dsl.common.extensions.EObjectPropertiesAdapter;
import org.openxma.dsl.core.model.AtomicBoolExpr;
import org.openxma.dsl.core.model.AtomicExpr;
import org.openxma.dsl.core.model.Call;
import org.openxma.dsl.core.model.CondORExpr;
import org.openxma.dsl.core.model.EqualityExpr;
import org.openxma.dsl.core.model.FalseLiteral;
import org.openxma.dsl.core.model.MultiplicativeExpr;
import org.openxma.dsl.core.model.RelationalExpr;
import org.openxma.dsl.core.model.TrueLiteral;
import org.openxma.dsl.core.util.NextOperatorSwitch;
import org.openxma.dsl.dom.model.Property;
import org.openxma.dsl.pom.model.CheckBox;
import org.openxma.dsl.pom.model.DataObjectVariable;
import org.openxma.dsl.pom.model.GuiElement;
import org.openxma.dsl.pom.model.ObjectProperty;
import org.openxma.dsl.pom.model.Page;
import org.openxma.xmadsl.GuidesignExtension;

import at.spardat.xma.guidesign.HiddenWidget;
import at.spardat.xma.guidesign.XMACompProperty;
import at.spardat.xma.guidesign.XMAComponent;
import at.spardat.xma.guidesign.XMAPage;
import at.spardat.xma.guidesign.XMAWidget;
import at.spardat.xma.guidesign.flex.ApplicationFunction;
import at.spardat.xma.guidesign.flex.ApplicationProperty;
import at.spardat.xma.guidesign.flex.Expression;
import at.spardat.xma.guidesign.flex.FlexFactory;
import at.spardat.xma.guidesign.flex.IExpVariable;
import at.spardat.xma.guidesign.flex.VariableExpression;
import at.spardat.xma.guidesign.flex.XMAApplicationContext;
import at.spardat.xma.guidesign.flex.type.VariableAccess;
import at.spardat.xma.guidesign.types.PropertyType;

/**
 * Xtend java extensions for core Expression.
 */
public class ExpressionExtension {
	private static final String NULL_LITERAL = "NULL";
	private static final String OPERATOR_STACK = "operator_stack";

	/**
	 * @param eObject
	 * @return the next available operator based on the given object
	 */
	public static String getNextOperator(EObject eObject) {
		return new NextOperatorSwitch().doSwitch(eObject);
	}
	
	/**
	 * Pushes the specified operator onto the top of the stack stored on the root container of the given object.
	 * 
	 * @param eObject the adapter holding the stack
	 * @param operator the operator to push
	 */
	public static void pushOperator(EObject eObject, String operator) {
		EObject rootContainer = EcoreUtil.getRootContainer(eObject);
		Stack<String> stack = (Stack<String>) EObjectPropertiesAdapter.getProperty(rootContainer, OPERATOR_STACK);
		if (stack==null) {
			stack = new Stack<String>();
			EObjectPropertiesAdapter.setProperty(rootContainer, OPERATOR_STACK,stack);
		}
		if (operator!=null && !NULL_LITERAL.equals(operator)) {
			stack.push(operator);
		}
	}
	
	/**
	 * @param eObject the adapter holding the stack
	 * @return the operator at the top of the stack and removes it
	 */
	public static String popOperator(EObject eObject) {
		EObject rootContainer = EcoreUtil.getRootContainer(eObject);
		Stack<String> stack = (Stack<String>) EObjectPropertiesAdapter.getProperty(rootContainer, OPERATOR_STACK);
		if (stack!=null && !stack.isEmpty()) {
			return stack.pop();
		}
		return "";
	}
	
	/**
	 * @param equalityExpr the expression to evaluate
	 * @return true of false whether the given expression evals to a primitive wrapper type
	 */
	public static boolean isPrimitiveWrapperType(EqualityExpr equalityExpr) {
		String typeName = new ExpressionTypeSwitch().doSwitch(equalityExpr);
		Boolean primitiveWrapperType = SimpleTypeExtension
				.isPrimitiveWrapperType(typeName);
		return primitiveWrapperType;
	}
		
	public static ApplicationFunction findApplicationFunction(Call call, XMAComponent xmaComponent) {

		XMAApplicationContext application = GuidesignExtension.getXMAApplicationContext(xmaComponent);
		
		for (Object obj : application.getFunctions()) {
			ApplicationFunction function = (ApplicationFunction)obj;
			if (function.getNamFunction().equals(call.getFunction().getName())) {
				return function;
			}
		}
		return createApplicationFunction(application,call,xmaComponent);		
	}
	
	public static at.spardat.xma.guidesign.flex.type.VariableAccess translateAccessType(EObject object, org.openxma.dsl.core.model.VariableAccess accessType) {		
		switch(accessType.getValue()) {
			case org.openxma.dsl.core.model.VariableAccess.VALUE_VALUE: return at.spardat.xma.guidesign.flex.type.VariableAccess.VALUE_LITERAL;
			case org.openxma.dsl.core.model.VariableAccess.HAS_VALUE_VALUE: return at.spardat.xma.guidesign.flex.type.VariableAccess.HAS_VALUE_LITERAL;
			case org.openxma.dsl.core.model.VariableAccess.GET_SELECTED_VALUE: return at.spardat.xma.guidesign.flex.type.VariableAccess.GET_SELECTED_LITERAL;
			case org.openxma.dsl.core.model.VariableAccess.IS_SELECTED_VALUE: return at.spardat.xma.guidesign.flex.type.VariableAccess.IS_SELECTED_LITERAL;
			case org.openxma.dsl.core.model.VariableAccess.GET_SELECTION_COUNT_VALUE: return at.spardat.xma.guidesign.flex.type.VariableAccess.GET_SELECTION_COUNT_LITERAL;
			case org.openxma.dsl.core.model.VariableAccess.SIZE_VALUE: return at.spardat.xma.guidesign.flex.type.VariableAccess.SIZE_LITERAL;
			case org.openxma.dsl.core.model.VariableAccess.IS_VALID_VALUE: return at.spardat.xma.guidesign.flex.type.VariableAccess.IS_VALID_LITERAL;
			
			default:
				if (object instanceof GuiElement) {
					if (object instanceof org.openxma.dsl.pom.model.RadioButton || object instanceof CheckBox) {
						return at.spardat.xma.guidesign.flex.type.VariableAccess.IS_SELECTED_LITERAL;					
					} 
				} 
				if (object instanceof XMAWidget) {
					if (object instanceof at.spardat.xma.guidesign.RadioButton || object instanceof at.spardat.xma.guidesign.CheckButton) {
						return at.spardat.xma.guidesign.flex.type.VariableAccess.IS_SELECTED_LITERAL;
					}
				} 
				return at.spardat.xma.guidesign.flex.type.VariableAccess.VALUE_LITERAL;							
		}
	}
	
	/*
	public static VariableExpression createXMAVariable(Table table, VariableAccess accessType, XMAComponent comp, XMAPage currPage, DataObjectVariable dov) {
		VariableExpression tableVariable = FlexFactory.eINSTANCE.createVariableExpression();
		List<XMAWidget> generatedWidgets = ComplexTypeExtension.getGeneratedXMAWidgetsForGuiElement(table);		
		if (generatedWidgets == null || generatedWidgets.size() == 0) {
			throw new RuntimeException("The generated XMAWidgets can not be determined for the GuiElement: "+table);
		}		
		XMATable xmaTable = (XMATable)generatedWidgets.get(0);		
		tableVariable.setTypAccess(accessType);
		tableVariable.setSource(xmaTable);	
		return tableVariable;
	}
	
	public static VariableExpression createXMAVariable(Text text, VariableAccess accessType, XMAComponent comp, XMAPage currPage, DataObjectVariable dov) {
		VariableExpression textVariable = FlexFactory.eINSTANCE.createVariableExpression();
		List<XMAWidget> generatedWidgets = ComplexTypeExtension.getGeneratedXMAWidgetsForGuiElement(text);		
		if (generatedWidgets == null || generatedWidgets.size() == 0) {
			throw new RuntimeException("The generated XMAWidgets can not be determined for the GuiElement: "+text);
		}		
		XMAText xmaText = (XMAText)generatedWidgets.get(0);		
		textVariable.setTypAccess(accessType);
		textVariable.setSource(xmaText);	
		return textVariable;
	}
	
	public static VariableExpression createXMAVariable(Combo combo, VariableAccess accessType, XMAComponent comp, XMAPage currPage, DataObjectVariable dov) {
		VariableExpression textVariable = FlexFactory.eINSTANCE.createVariableExpression();
		List<XMAWidget> generatedWidgets = ComplexTypeExtension.getGeneratedXMAWidgetsForGuiElement(combo);		
		if (generatedWidgets == null || generatedWidgets.size() == 0) {
			throw new RuntimeException("The generated XMAWidgets can not be determined for the GuiElement: "+combo);
		}		
		SimpleCombo simpleCombo = (SimpleCombo)generatedWidgets.get(0);		
		textVariable.setTypAccess(accessType);
		textVariable.setSource(simpleCombo);	
		return textVariable;
	}	
	*/
		
	public static VariableExpression createXMAVariableForOtherGuiElement(GuiElement guiElement, VariableAccess accessType, XMAComponent comp, XMAPage currPage, DataObjectVariable dov) {
		VariableExpression textVariable = FlexFactory.eINSTANCE.createVariableExpression();
		List<XMAWidget> generatedWidgets = ComplexTypeExtension.getGeneratedXMAWidgetsForGuiElement(guiElement);		
		if (generatedWidgets == null || generatedWidgets.size() == 0) {
			throw new RuntimeException("The generated XMAWidget can not be determined for the GuiElement: "+guiElement);
		}		
		IExpVariable iExpVariable = (IExpVariable)generatedWidgets.get(0);		
		textVariable.setTypAccess(accessType);
		textVariable.setSource(iExpVariable);	
		return textVariable;
	}			
	
	public static VariableExpression createXMAVariable(Page page, VariableAccess accessType, XMAComponent comp, XMAPage currPage, DataObjectVariable dov) {
		VariableExpression pageVariable = FlexFactory.eINSTANCE.createVariableExpression();			
		XMAPage referencedXMAPage = GuidesignExtension.findXMAPage(comp,page);				
		pageVariable.setTypAccess(accessType);
		pageVariable.setSource(referencedXMAPage);	
		return pageVariable;
	}			
	
	/*
	 * Statusflag - variable
	 */ 
	// TODO use in Extend
	public static VariableExpression createXMAVariable(Property property, VariableAccess accessType, XMAComponent comp, XMAPage currPage, DataObjectVariable dov) {

		VariableExpression stateFlagVariable = FlexFactory.eINSTANCE.createVariableExpression();		
		ApplicationProperty referencedProperty = findApplicationProperty(property,comp,dov);
			
		stateFlagVariable.setTypAccess(accessType);	
		stateFlagVariable.setSource(referencedProperty);
		
		return stateFlagVariable; 	
	}
	
	
	public static VariableExpression createXMAVariable(ObjectProperty property, VariableAccess accessType, XMAComponent comp, XMAPage currPage, DataObjectVariable dov) {

		VariableExpression stateFlagVariable = FlexFactory.eINSTANCE.createVariableExpression();		
		IExpVariable expVariable = findHiddenWidget(currPage,property.getName());
		if (expVariable == null) {
			expVariable = findXMACompProperty(comp,property.getName());			
		}
			
		stateFlagVariable.setTypAccess(accessType);	
		stateFlagVariable.setSource(expVariable);
		
		return stateFlagVariable; 	
	}
	
	private static HiddenWidget findHiddenWidget(XMAPage currPage, String name) {
		if (currPage.getComposite() != null) {
			for (EObject obj : currPage.getComposite().eContents()) {
				if (obj instanceof HiddenWidget && ((HiddenWidget)obj).getNamInstance().equals(name)) {
					return (HiddenWidget)obj;
				}
			}		
		}
		return null;		
	}
	
	private static XMACompProperty findXMACompProperty(XMAComponent comp, String name) {
		for (Object obj : comp.getProperty()) {
			XMACompProperty cp = (XMACompProperty)obj;
			if (cp.getNamInstance().equals(name)) {
				return  cp;
			}
		}
		return null;
	}
	
	// TODO use in Extend (first check and implement used functions)
//	public static XMAStateFlag findXMAStateFlag(StatusFlag statusFlag, XMAComponent comp, XMAPage currPage, DataObjectVariable dov ) {
//
//		EObject stateFlagParent = StatusFlagExtension.getOwnerOfStatusFlag(statusFlag);
//
//		if (stateFlagParent instanceof Component) {
//			return StatusFlagExtension.findXMAStateFlagForComponent(comp,statusFlag,dov);
//		} else if (stateFlagParent instanceof Page || stateFlagParent instanceof Entity) {
//			return StatusFlagExtension.findXMAStateFlagForPage(currPage,statusFlag,comp,dov);
//		} else if (stateFlagParent instanceof ApplicationSession) {
//			return StatusFlagExtension.findXMAStateFlagForApplication(comp,statusFlag,dov);
//		}
//		throw new RuntimeException("XMAStateFlag not found "+statusFlag.getName());
//	}
	

	// TODO use in Extend
	public static ApplicationProperty findApplicationProperty(Property property, XMAComponent xmaComponent, DataObjectVariable dov) {
		XMAApplicationContext application = GuidesignExtension.getXMAApplicationContext(xmaComponent);
		String propName = property.getName();		
		for (Object obj : application.getProperties()) {
			ApplicationProperty appProp = (ApplicationProperty)obj;
			if (appProp.getNamProperty().equals(propName)) {
				return appProp;
			}
		}
		return createApplicationProperty(application,property,xmaComponent);
	}
	
	/**
	 * Mapping for supported property-types
	 */		
	public static PropertyType getXMAPropertyType(Property property) {
		// TODO replace this dummy implementation with something else which
		// allows a save and consistent type determination
		String typeName = property.getType().getName();		
		if ("String".equals(typeName)) return PropertyType.STRING_LITERAL;
		if ("Boolean".equals(typeName))  return PropertyType.BOOLEAN_LITERAL;
		if ("Date".equals(typeName))  return PropertyType.DATE_LITERAL;
		if ("Integer".equals(typeName))  return PropertyType.INT_LITERAL;
		if ("Number".equals(typeName))  return PropertyType.DOUBLE_LITERAL;		// TODO: discuss rename to "Decimal"	
		return PropertyType.STRING_LITERAL;	
	}
	
	public static ApplicationProperty createApplicationProperty(XMAApplicationContext application, Property property, XMAComponent xmaComponent) {
		
		ApplicationProperty appProperty = FlexFactory.eINSTANCE.createApplicationProperty();
		
		appProperty.setTypProperty(getXMAPropertyType(property));
		appProperty.setNamProperty(property.getName());
		
		application.getProperties().add(appProperty);
			
		return appProperty;	
	}
		
	public static ApplicationFunction createApplicationFunction(XMAApplicationContext application, Call call, XMAComponent xmaComponent) {		
		ApplicationFunction appFunction = FlexFactory.eINSTANCE.createApplicationFunction();	
		appFunction.setNamFunction(call.getFunction().getName());		
		application.getFunctions().add(appFunction);			
		return appFunction;
	}
	
	public static Expression cloneExpression(Expression exp) {
		return (Expression)EcoreUtil.copy(exp);
	}
	
	/**
	 * Determine the boolean literal of a EqualityExpr.
	 * In case the expression is not a boolean literal, the
	 * method returns null.
	 * 
	 * @param expression
	 * @return
	 */
	public static Boolean booleanLiteralFor(EqualityExpr expression) {
		if (expression.getRight() == null) {
			CondORExpr orExpr = expression.getLeft();
			if (orExpr.getRights().size() == 0 && orExpr.getLeft().getRights().size() == 0) {
				AtomicBoolExpr boolExpr = orExpr.getLeft().getLeft();
				if (boolExpr instanceof RelationalExpr) {
					RelationalExpr re = (RelationalExpr)boolExpr;
					if (re.getRight() == null && re.getLeft().getRights().size() == 0) {
						MultiplicativeExpr me = re.getLeft().getLeft();
						if (me.getRights().size() == 0) {
							AtomicExpr ae = me.getLeft();
							if (ae instanceof TrueLiteral) {
								return true;
							} else if (ae instanceof FalseLiteral) {
								return false;
							} else {
								return null;
							}
						}
					}
				}					
			}
		}
		return null;
	}	

}
