package org.openxma.dsl.generator.helper;

import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.openxma.dsl.common.extensions.EObjectPropertiesAdapter;
import org.openxma.dsl.core.model.BoolLiteral;
import org.openxma.dsl.core.model.Style;
import org.openxma.dsl.core.model.TrueLiteral;
import org.openxma.dsl.core.validation.ValidationCollectionUtil;
import org.openxma.dsl.generator.GeneratorProperties;
import org.openxma.dsl.generator.impl.DefaultLayoutStrategy;
import org.openxma.dsl.pom.PomFactory;
import org.openxma.dsl.pom.model.AbsoluteSiblingAttachment;
import org.openxma.dsl.pom.model.AttachmentOwnPosition;
import org.openxma.dsl.pom.model.AttachmentProperty;
import org.openxma.dsl.pom.model.AttachmentSiblingPosition;
import org.openxma.dsl.pom.model.AttachmentSpecification;
import org.openxma.dsl.pom.model.ComplexElement;
import org.openxma.dsl.pom.model.ComposeData;
import org.openxma.dsl.pom.model.Composite;
import org.openxma.dsl.pom.model.Content;
import org.openxma.dsl.pom.model.CustomizeableField;
import org.openxma.dsl.pom.model.FieldPart;
import org.openxma.dsl.pom.model.FieldPartSpecification;
import org.openxma.dsl.pom.model.GuiElement;
import org.openxma.dsl.pom.model.HeightProperty;
import org.openxma.dsl.pom.model.IAttachmentPosition;
import org.openxma.dsl.pom.model.IComposite;
import org.openxma.dsl.pom.model.IElementOnWhichCanBeAttached;
import org.openxma.dsl.pom.model.IElementWithLayoutData;
import org.openxma.dsl.pom.model.ITabPage;
import org.openxma.dsl.pom.model.LayoutDataProperty;
import org.openxma.dsl.pom.model.NoneAttachment;
import org.openxma.dsl.pom.model.ParentAttachment;
import org.openxma.dsl.pom.model.RelativeSibling;
import org.openxma.dsl.pom.model.RelativeSiblingAttachment;
import org.openxma.dsl.pom.model.SetOfGuiElements;
import org.openxma.dsl.pom.model.SiblingAttachment;
import org.openxma.dsl.pom.model.Table;
import org.openxma.dsl.pom.model.TabulatorAttachment;
import org.openxma.dsl.pom.model.TabulatorPosition;
import org.openxma.dsl.pom.model.WidthProperty;
import org.openxma.dsl.pom.model.XmadslComposite;
import org.openxma.dsl.pom.naming.PomNameProvider;
import org.openxma.dsl.pom.util.LayoutData;
import org.openxma.dsl.pom.validation.PomDslJavaValidator;

import at.spardat.xma.guidesign.CheckButton;
import at.spardat.xma.guidesign.GuidesignFactory;
import at.spardat.xma.guidesign.RadioButton;
import at.spardat.xma.guidesign.XMAButton;
import at.spardat.xma.guidesign.XMAComposite;
import at.spardat.xma.guidesign.XMAFormAttachment;
import at.spardat.xma.guidesign.XMAFormData;
import at.spardat.xma.guidesign.XMAGroup;
import at.spardat.xma.guidesign.XMASeperator;
import at.spardat.xma.guidesign.XMATable;
import at.spardat.xma.guidesign.XMATree;
import at.spardat.xma.guidesign.XMAWidget;
import at.spardat.xma.guidesign.types.AttachSideType;


public class FormDataExtension {

	public static final String LAST_TABULATOR_POSITION_PROPERTY_NAME = "lastTabulatorPosition";

	public static LayoutData getLayoutData(GuiElement guiElement) {

		if (guiElement instanceof IElementWithLayoutData) {
			return new LayoutData((IElementWithLayoutData)guiElement);
		}
		return null;
/*
		LayoutData layoutData = null;
		if (guiElement instanceof SimpleElement) {
			SimpleElement simpleElement = (SimpleElement)guiElement;
			layoutData = simpleElement.getLayoutData();
		} else if (guiElement instanceof Table) {
			Table table = (Table)guiElement;
			layoutData = table.getLayoutData();
		} else if (guiElement instanceof XmadslComposite) {
			XmadslComposite comp = (XmadslComposite)guiElement;
			layoutData = comp.getLayoutData();
		}
		return layoutData;
		*/
	}

	public static void setWidgetDimension(XMAFormData formData, GuiElement guiElement, XMAWidget widget, FieldPartSpecification fieldPartSpecification) {

		 FieldPart fieldPart = WidgetExtension.getFieldPartProperty(widget);

		// Do only use dimension of CustomizeableField for Control (not label or unitLabel)
		if (fieldPart == null || FieldPart.CONTROL.equals(fieldPart)) {			
			// Dimension from layoutData
			LayoutData layoutData = getLayoutData(guiElement);
			if (layoutData != null) {
				// Width
				if (layoutData.getWidth()>0) {
					formData.setQntWidth(layoutData.getWidth());
				}
				// Height
				if (layoutData.getHeight()>0) {
					formData.setQntHeight(layoutData.getHeight());
				}
			}
		}
		
		// Dimension from fieldPartSpecification
		if (fieldPartSpecification != null) {
			for (LayoutDataProperty p : fieldPartSpecification.getLayoutProperties()) {
				if (p instanceof WidthProperty) {
					formData.setQntWidth(((WidthProperty)p).getWidth());
				}
				if (p instanceof HeightProperty) {
					formData.setQntHeight(((HeightProperty)p).getHeight());
				}			
			}
		}
	}

	public static ComposeData getComposeData(Composite composite) {

		if (composite instanceof IComposite) {
			Content content =  ((IComposite)composite).getContent();
			return content != null ? content.getComposeLayout() : null;
		}
		if (composite instanceof ITabPage) {
			return ((ITabPage)composite).getComposeLayout();
		}
		return null;
	}

	public static void setComposeData(Composite composite, ComposeData composeData) {

		if (composite instanceof IComposite && composeData != null) {
			if (((IComposite)composite).getContent()==null) {
				((IComposite)composite).setContent(PomFactory.eINSTANCE.createContent());
			}
			((IComposite)composite).getContent().setComposeLayout(composeData);
			return;
		}
		if (composite instanceof ITabPage && composeData != null) {
			((ITabPage)composite).setComposeLayout(composeData);
			return;
		}
	}

	public static boolean hasComposeData(Composite composite) {
		if (composite instanceof IComposite) {
            return true;
        }
		if (composite instanceof ITabPage) {
            return true;
        }
		return false;
	}

	/**
	 * Get the xmadsl attachment for the specified own position. In case no
	 * has been found for left or top try to find center attachments:
	 * .) hcenter for left
	 * .) vcenter for top
	 *
	 * (In XMA the center attachments can be attached to the own top or left
	 * attachment.)
	 */
	public static AttachmentSpecification getAttachment(LayoutData layoutData, final AttachmentOwnPosition ownPosition) {
		return LayoutData.getAttachment(layoutData, ownPosition);
	}

	/**
	 * Get the XMA attach side for the sibling. In case the sibling attach side is not defined
	 * it is determined by the own attachment side.
	 */
	public static AttachSideType getXMAAttachSide(AttachmentSiblingPosition position, AttachmentOwnPosition ownPosition) {

		// Is the sibling attach side defined?
		if (position != null && position != AttachmentSiblingPosition.NULL) {
			switch (position.getValue()) {
				case AttachmentSiblingPosition.TOP_VALUE: return AttachSideType.TOP_LITERAL;
				case AttachmentSiblingPosition.BOTTOM_VALUE: return AttachSideType.BOTTOM_LITERAL;
				case AttachmentSiblingPosition.LEFT_VALUE: return AttachSideType.LEFT_LITERAL;
				case AttachmentSiblingPosition.RIGHT_VALUE: return AttachSideType.RIGHT_LITERAL;
				case AttachmentSiblingPosition.CENTER_VALUE: return AttachSideType.CENTER_LITERAL;
				default: throw new RuntimeException("Sibling attach position not supported, position/ownPosition: "+position+"/"+ownPosition);
			 }
		} else {
			switch (ownPosition.getValue()) {
				case AttachmentOwnPosition.TOP_VALUE: return AttachSideType.BOTTOM_LITERAL;
				case AttachmentOwnPosition.BOTTOM_VALUE: return AttachSideType.TOP_LITERAL;
				case AttachmentOwnPosition.LEFT_VALUE: return AttachSideType.RIGHT_LITERAL;
				case AttachmentOwnPosition.RIGHT_VALUE: return AttachSideType.LEFT_LITERAL;
				case AttachmentOwnPosition.VCENTER_VALUE: return AttachSideType.CENTER_LITERAL;
				case AttachmentOwnPosition.HCENTER_VALUE: return AttachSideType.CENTER_LITERAL;
				default: throw new RuntimeException("Own attach position not supported, position/ownPosition: "+position+"/"+ownPosition);
			}
		}
	}


	public static XMAFormAttachment createXMAAttachment(final AttachmentSpecification attachmentSpec, XMAComposite parentComposite, GuiElement prevGuiElement, GuiElement nextGuiElement, AttachmentOwnPosition ownPosition) {
		if (attachmentSpec instanceof NoneAttachment) {
			return null;
		}
		if (attachmentSpec instanceof SiblingAttachment) {
			return createXMAAttachment((SiblingAttachment)attachmentSpec,parentComposite,prevGuiElement,nextGuiElement,ownPosition);
		}
		if (attachmentSpec instanceof ParentAttachment) {
			return createXMAAttachment((ParentAttachment)attachmentSpec,parentComposite,prevGuiElement,nextGuiElement,ownPosition);
		}
		if (attachmentSpec instanceof TabulatorAttachment) {
			return createXMAAttachment((TabulatorAttachment)attachmentSpec,parentComposite,prevGuiElement,nextGuiElement,ownPosition);
		}
		throw new RuntimeException("Unknown type of AttachmentSpecification: "+attachmentSpec);
	}

	/**
	 * Create a XMAFormAttachment from a  ParentAttachment
	 */
	public static XMAFormAttachment createXMAAttachment(final ParentAttachment parentAttachment, XMAComposite parentComposite, GuiElement prevGuiElement, GuiElement nextGuiElement, AttachmentOwnPosition ownPosition) {
		try {
			XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
			if (parentAttachment.isRelative()) {
				formAttachment.setQntNominator(parentAttachment.getDistance());
				if (parentAttachment.getOffset() != null) {
					formAttachment.setQntOffset(parentAttachment.getOffset().getSignedValue());
				}
			} else {
				formAttachment.setQntOffset(parentAttachment.getDistance());
			}
			return formAttachment;
		} catch (RuntimeException ex) {
			ex.printStackTrace();
			throw ex;
		}
	}
	
	/**
	 * Create a XMAFormAttachment from a  ParentAttachment
	 */
	public static XMAFormAttachment createXMAAttachment(final NoneAttachment parentAttachment, XMAComposite parentComposite, GuiElement prevGuiElement, GuiElement nextGuiElement, AttachmentOwnPosition ownPosition) {
		return null;
	}

	public static IAttachmentPosition getBaseIAttachmentPosition(IAttachmentPosition iAttachmentPosition) {
		if (iAttachmentPosition != null && iAttachmentPosition.getReferencedPosition() != null) {
			return getBaseIAttachmentPosition(iAttachmentPosition.getReferencedPosition());
		} else {
			return iAttachmentPosition;
		}
	}

	/**
	 * Create a XMAFormAttachment from a SiblingAttachment
	 */
	public static XMAFormAttachment createXMAAttachment(final SiblingAttachment attachment, XMAComposite parentComposite, GuiElement prevGuiElement, GuiElement nextGuiElement,AttachmentOwnPosition ownPosition) {
		try {
			
			IElementOnWhichCanBeAttached siblingElement = null;			
			XMAWidget sibling = null;					
			 
			if (attachment instanceof AbsoluteSiblingAttachment) {				
				AbsoluteSiblingAttachment asa = (AbsoluteSiblingAttachment)attachment;
				
				// Sibling is a TabulatorPosition				
				if (asa.getSibling() instanceof IAttachmentPosition) {
					IAttachmentPosition attachmentPosition = getBaseIAttachmentPosition((IAttachmentPosition)asa.getSibling());
					XMAFormAttachment formAttachment = createXMAAttachment(getParentSetOfSimpleElements(attachment),attachmentPosition);
					if (attachment.getOffset() != null) {
						formAttachment.setQntOffset(formAttachment.getQntOffset()+attachment.getOffset().getSignedValue());
					}
					return formAttachment;
				}
				
				// Sibling is a FieldPart
				if (asa.getFieldPart() != null && asa.getSibling() == null) {
					FieldPartSpecification fps = PomDslJavaValidator.getFieldPartSpecificationContainer(attachment);
					CustomizeableField f = (CustomizeableField)fps.eContainer();
					FieldPart fp = asa.getFieldPart();
					
					for (Object obj : parentComposite.getControls()) {
						if (obj instanceof XMAWidget) {
							XMAWidget w = (XMAWidget)obj;
							GuiElement guiElement = WidgetExtension.getGuiElementProperty(w);
							if (guiElement == f) {
								FieldPartSpecification fieldPartSpecification = WidgetExtension.getFieldPartSpecificationProperty(w);
								if (fieldPartSpecification.getPart().getValue() == fp.getValue()) {
									sibling = w;
									break;
								}
							}												
						}
					}
					
				// Sibling is GuiElement	
				} else {
					siblingElement = asa.getSibling();
				}
								
			} else if (attachment instanceof RelativeSiblingAttachment) {
				RelativeSiblingAttachment rsa = (RelativeSiblingAttachment)attachment;					
				switch (rsa.getRelativeSibling().getValue()) {
				case RelativeSibling.PREVIOUS_VALUE:
					siblingElement = (IElementOnWhichCanBeAttached)prevGuiElement;
					break;
				case RelativeSibling.NEXT_VALUE:
					siblingElement = (IElementOnWhichCanBeAttached)nextGuiElement;
					break;
				}				
			}

			// Determine according sibling XMAWidget			
			if (sibling == null && siblingElement != null) {				

				String siblingId = PomNameProvider.getXmaName(siblingElement);
				if (siblingId == null) {
					siblingId = PomNameProvider.composeDefaultNameForGuiElement(siblingElement);
				}
				if (siblingElement instanceof XmadslComposite) {
					BoolLiteral scroll = ((XmadslComposite)siblingElement).getScrollable();
					if (scroll instanceof TrueLiteral) {
						siblingId=PomNameProvider.getScrolledName(siblingId);
					}					
				}
				if (siblingElement instanceof Table && ((Table)siblingElement).getPageable() instanceof TrueLiteral) {
					siblingId = PomNameProvider.getTableCompositeName(siblingId);	
				}					
				sibling = ValidationCollectionUtil.findFirstElementWithId(parentComposite.eContents(), XMAWidget.class,
						siblingId, PomNameProvider.XMA_ID);				

			}
			
		

			// Determine attach side
			AttachSideType codAttachSide = getXMAAttachSide(attachment.getSiblingPosition(),ownPosition);

			// Create FormAttachment
			XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
			formAttachment.setCodAttachSide(codAttachSide);
			formAttachment.setAttachElement(sibling);
			if (attachment.getOffset() != null) {
				formAttachment.setQntOffset(attachment.getOffset().getSignedValue());
			} else {
				formAttachment.setQntOffset(0);
			}
			return formAttachment;
		} catch (RuntimeException ex) {
			ex.printStackTrace();
			throw ex;
		}
	}



	private static SetOfGuiElements getParentSetOfSimpleElements(AttachmentSpecification attachmentSpec) {
		AttachmentProperty attachment = (AttachmentProperty)attachmentSpec.eContainer();
		if (attachment != null) {
			GuiElement guiElement = null;			
			if (attachment.eContainer() instanceof FieldPartSpecification) {
				FieldPartSpecification fps = (FieldPartSpecification)attachment.eContainer();
				guiElement = (GuiElement)fps.eContainer();				
				
			} else {
				guiElement = (GuiElement)attachment.eContainer();
			}

			if (guiElement != null && guiElement.eContainer() instanceof SetOfGuiElements) {
				SetOfGuiElements setOfSimpleElements	= (SetOfGuiElements)guiElement.eContainer();
				return setOfSimpleElements;
			} else {
				return null;
			}

		}
		return null;
	}

	/**
	 * Create a XMAFormAttachment from a TabulatorAttachment
	 */
	public static XMAFormAttachment createXMAAttachment(TabulatorAttachment tabulatorAttachment, XMAComposite parentComposite, GuiElement prevGuiElement, GuiElement nextGuiElement, AttachmentOwnPosition ownPosition) {
		try {
			GuiElement guiElement = WidgetExtension.getGuiElementProperty(parentComposite);
			// Additional XMAComposite nested in case of SetOfSimpleElement
			if (guiElement == null) {
				EObject sourceModelElement = WidgetExtension.getSourceModelElementProperty(parentComposite);
				if (sourceModelElement instanceof SetOfGuiElements) {
					guiElement = WidgetExtension.getGuiElementProperty((XMAWidget)parentComposite.eContainer());
				}
			}
			if (guiElement instanceof Composite) {
				Composite composite = (Composite)guiElement;
				ComposeData composeData = getComposeData(composite);
				TabulatorPosition tabulatorPosition = getTabulatorPosition(composeData,tabulatorAttachment.getTabIndex()-1);
				IAttachmentPosition attachmentPosition = getBaseIAttachmentPosition(tabulatorPosition);
				if (attachmentPosition != null) {
					XMAFormAttachment formAttachment = createXMAAttachment(getParentSetOfSimpleElements(tabulatorAttachment),attachmentPosition);
					if (tabulatorAttachment.getOffset() != null) {
						formAttachment.setQntOffset(formAttachment.getQntOffset()+tabulatorAttachment.getOffset().getSignedValue());
					}
					return formAttachment;
				}
			}
			return null;
		} catch (RuntimeException ex) {
			ex.printStackTrace();
			throw ex;
		}
	}

	/**
	 * Create a XMAFormAttachment from a TabulatorPosition
	 */
	public static XMAFormAttachment createXMAAttachment(SetOfGuiElements parentSetOfSimpleElements, IAttachmentPosition tabulatorPosition) {
		XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
 		if (tabulatorPosition.isRelative()) {
 			formAttachment.setQntNominator(tabulatorPosition.getWidth());
 			if (tabulatorPosition.getOffset() != null) {
 				formAttachment.setQntOffset(tabulatorPosition.getOffset().getSignedValue());
 			}
 		} else {
 			formAttachment.setQntOffset(tabulatorPosition.getWidth());
 		}
 		if (parentSetOfSimpleElements != null && tabulatorPosition instanceof TabulatorPosition) {
 			setLastTabulatorPosition(parentSetOfSimpleElements,(TabulatorPosition)tabulatorPosition);
 		}
		return formAttachment;
	}

	public static TabulatorPosition getTabulatorPosition(ComposeData composeData, Integer index) {
		if (index == null || index < 0) {
            return null;
        }
		if (composeData != null && composeData.getTabulators() != null && composeData.getTabulators().size()>index) {
			return composeData.getTabulators().get(index);
		}
		return null;
	}

	private static int getTabulatorPositionIndex(TabulatorPosition tab, ComposeData composeData) {
		if (tab != null && composeData != null && composeData.getTabulators() != null) {
			return composeData.getTabulators().indexOf(tab);
		}
		return -1;
	}

	public static TabulatorPosition getLeftTabulatorPosition(XMAWidget xmaWidget, GuiElement guiElement, SetOfGuiElements parentSetOfSimpleElements, ComposeData composeData, String columnIndex, Boolean isLabel, Boolean isMainWidget) {
		return getLeftTabulatorPosition(xmaWidget,guiElement,parentSetOfSimpleElements,composeData,Integer.parseInt(columnIndex),isLabel,isMainWidget);
	}
	public static TabulatorPosition getLeftTabulatorPosition(XMAWidget xmaWidget, GuiElement guiElement, SetOfGuiElements parentSetOfSimpleElements, ComposeData composeData, int columnIndex, Boolean isLabel, Boolean isMainWidget) {

		TabulatorPosition tabLeft = null;
		boolean useTabInDefault = useLeftTabInDefaultAttachment(xmaWidget,guiElement,isLabel,isMainWidget);
		if (useTabInDefault) {
			tabLeft = getLastTabulatorPosition(parentSetOfSimpleElements);
			if (tabLeft == null && composeData != null) {
				tabLeft = getTabulatorPosition(composeData,columnIndex-1);
			}
		}
		setLastTabulatorPosition(parentSetOfSimpleElements,tabLeft);
		return tabLeft;
	}

	public static TabulatorPosition getRightTabulatorPosition(XMAWidget xmaWidget, GuiElement guiElement, SetOfGuiElements parentSetOfSimpleElements, ComposeData composeData, String columnIndex, Boolean isLabel, Boolean isMainWidget) {
		return getRightTabulatorPosition(xmaWidget,guiElement,parentSetOfSimpleElements,composeData,Integer.parseInt(columnIndex),isLabel,isMainWidget);
	}

	public static TabulatorPosition getRightTabulatorPosition(XMAWidget xmaWidget, GuiElement guiElement, SetOfGuiElements parentSetOfSimpleElements, ComposeData composeData, int columnIndex, Boolean isLabel, Boolean isMainWidget) {

		TabulatorPosition tabRight = null;
		boolean useTabInDefault = useRightTabInDefaultAttachment(xmaWidget,guiElement,isLabel,isMainWidget);

			TabulatorPosition lastTab = getLastTabulatorPosition(parentSetOfSimpleElements);
			int indexForLastTab = getTabulatorPositionIndex(lastTab,composeData);
			if (indexForLastTab>=0) {
				int span=1;
				if (xmaWidget instanceof CheckButton && GeneratorProperties.getLayoutStrategy().isCheckBoxLeftAligned()) {
					span=2;
				}
				tabRight = getTabulatorPosition(composeData,indexForLastTab+span);
			}
			if (tabRight == null && composeData != null) {
				int rightIndex = columnIndex;
				if (xmaWidget instanceof CheckButton && GeneratorProperties.getLayoutStrategy().isCheckBoxLeftAligned()) {
					rightIndex = columnIndex+1;
				}
				tabRight = getTabulatorPosition(composeData,rightIndex);
			}
			
			setLastTabulatorPosition(parentSetOfSimpleElements,tabRight);
			if (useTabInDefault) {			
				return tabRight;
			} 
			return null;
		}

	public static int getPositionOffset(IAttachmentPosition tab) {
		if (tab.isRelative()) {
			return tab.getOffset() != null ? tab.getOffset().getSignedValue() : 0;
		} else {
			return tab.getWidth();
		}
	}

	public static int getPositionNominator(IAttachmentPosition tab) {
		if (tab.isRelative()) {
			return tab.getWidth();
		} else {
			return 0;
		}
	}

	private static boolean isAttachedToWidget(XMAWidget widget, XMAFormAttachment attachment) {
		if (attachment != null && attachment.getAttachElement() != null) {
			return widget.getNamInstance().equals(attachment.getAttachElement().getNamInstance());
		}
		return false;
	}

	public static boolean attachNextWidgetsLeft(XMAWidget widget, XMAWidget nextWidget, XMAFormData formData) {
		int width = formData.getQntWidth();
		if (formData.getRightAttach() != null) {
			// This widget is attached to the next widget
			if (isAttachedToWidget(nextWidget,formData.getRightAttach())) {
                return false;
            }
			// This widget has an other right attachment
			return true;
		}
		// The right side of this widget is well defined
		if (formData.getLeftAttach() != null /*&& width != 0*/) {
            return true;
        }
		return false;
	}

	/**
	 * Determine whether a default attachment for left side should be used
	 * depending on other layout information:
	 *
	 * No default left attachment is required in case a right attachment and the
	 * width is defined, because the layout is defined without a left attachment.
	 * Furthermore a left attachment could even be inconsistent with this given
	 * information.
	 */
	public static boolean allowDefaultLeftAttachment(AttachmentSpecification rightAttachment, XMAFormData formData, XMAWidget widget, GuiElement guiElement, XMAWidget nextWidgetInSet, Boolean attachNextWidgetsLeft, Boolean isLabel, Boolean isMainWidget, Boolean tabulatorAvailable) {

		if (tabulatorAvailable != null && tabulatorAvailable.booleanValue() && useLeftTabInDefaultAttachment(widget,guiElement,isLabel,isMainWidget)) {
            return true;
        }
		if (attachNextWidgetsLeft != null && attachNextWidgetsLeft == true) {
			
			FieldPartSpecification fieldPartSpecification = WidgetExtension.getFieldPartSpecificationProperty(widget);			

			// Widget CAN be attached left, BUT has explicit defined right
			// attachment and width. This defines implicit the left position
			// and therefore no DEFAULT left attachment is made.
			GuiElement nextGuiElement = WidgetExtension.getGuiElementProperty(widget);
			if (nextGuiElement != null && hasRightAttachmentAndWidth(nextGuiElement,fieldPartSpecification)) {
				return false;
			}

			// Widget CAN be attached left and is not overruled by other
			// horizontal layout information
            return true;
        }

		// Left attachment is not possible because of the layout of the
		// previous widget.
		if (attachNextWidgetsLeft != null && attachNextWidgetsLeft == false) {
			return false;
		}

		if (rightAttachment == null) {
			return true;
		}
		return false;
	}

	private static boolean hasRightAttachmentAndWidth(GuiElement nextGuiElement, FieldPartSpecification fieldPartSpecification) {
		if (nextGuiElement != null) {
			AttachmentSpecification nextRightAttachment = LayoutDataExtension.getRightAttachment(nextGuiElement,fieldPartSpecification,true);
			Integer nextWidth = LayoutDataExtension.getWidth(nextGuiElement);
			if (nextRightAttachment != null && nextWidth != null && nextWidth>0) {
				return true;
			}
		}
		return false;
	}

	public static boolean allowDefaultRightAttachment(AttachmentSpecification leftAttachment, XMAFormData formData, XMAWidget widget, GuiElement guiElement, XMAWidget nextWidgetInSet, Boolean attachNextWidgetsLeft, Boolean isLabel, Boolean isMainWidget, Boolean tabulatorAvailable) {

		if (tabulatorAvailable && useRightTabInDefaultAttachment(widget,guiElement,isLabel,isMainWidget)) {
            return true;
        }
		if (attachNextWidgetsLeft != null && attachNextWidgetsLeft) {
            return false;
        }
		if (isButton(widget) && formData.getLeftAttach() == null) {
            return true;
        }
		if (widget instanceof XMASeperator) {
            return true;
        }
		if (guiElement instanceof ComplexElement) {
            return true;
        }
//		if (formData.getLeftAttach() == null || width==0) {
		if (formData.getLeftAttach() == null) {//next is not a button?
			return true;
		}
		return false;
	}

	public static boolean useLeftTabInDefaultAttachment(XMAWidget widget, GuiElement guiElement, Boolean isLabel, Boolean isMainWidget) {
		

		
		try {
			List<Style> styleList = PresentationModelExtension.getStyleProperty(guiElement);			
			FieldPart fieldPart = WidgetExtension.getFieldPartProperty(widget);
			if (styleList != null) {
				for (int i=styleList.size()-1;i>=0;i--) {
					Style style = styleList.get(i);
					Boolean flag = GeneratorProperties.getLayoutStrategy().useLeftTabInDefaultAttachment(widget,fieldPart,style.getName());
					if (flag != null) {
                        return flag.booleanValue();
                    }
				}
			}
			Boolean flag = GeneratorProperties.getLayoutStrategy().useLeftTabInDefaultAttachment(widget,fieldPart,null);
			if (flag != null) {
                return flag.booleanValue();
            }
			return DefaultLayoutStrategy.instance.useLeftTabInDefaultAttachment(widget,fieldPart,null);
		} catch (RuntimeException ex) {
			ex.printStackTrace();
			throw ex;
		}
	}

	public static boolean useRightTabInDefaultAttachment(XMAWidget widget, GuiElement guiElement, Boolean isLabel, Boolean isMainWidget) {
		try {
			List<Style> styleList = PresentationModelExtension.getStyleProperty(guiElement);		
            FieldPart fieldPart = WidgetExtension.getFieldPartProperty(widget);
			if (styleList != null) {
				for (int i=styleList.size()-1;i>=0;i--) {
					Style style = styleList.get(i);
					Boolean flag = GeneratorProperties.getLayoutStrategy().useRightTabInDefaultAttachment(widget,fieldPart,style.getName());
					if (flag != null) {
                        return flag.booleanValue();
                    }
				}
			}
			Boolean flag = GeneratorProperties.getLayoutStrategy().useRightTabInDefaultAttachment(widget,fieldPart,null);
			if (flag != null) {
                return flag.booleanValue();
            }
			return DefaultLayoutStrategy.instance.useRightTabInDefaultAttachment(widget,fieldPart,null);
		} catch (RuntimeException ex) {
			ex.printStackTrace();
			throw ex;
		}
	}

	/**
	 * Creates the default left attachment
	 *
	 * @param widget
	 * @param composeData
	 * @param tab
	 * @param isLabel
	 * @param isMainWidget
	 * @param prevWidgetInSet
	 * @param nextWidgetInSet
	 * @param attachNextWidgetsLeft2
	 * @return
	 */
	public static XMAFormAttachment createDefaultLeftAttachment(XMAWidget widget, ComposeData composeData, TabulatorPosition tab, Boolean isLabel, Boolean isMainWidget, XMAWidget prevWidgetInSet, XMAWidget nextWidgetInSet, Boolean attachNextWidgetsLeft2) {
		try {
			//int padding = composeData != null && composeData.getSpacing() != null ? composeData.getSpacing().getWidth() : 3;
			int padding = 3;
			// Attach to tabulator if available
            if (tab != null) {
    			return createAttachmentFromTabulatorPosition(tab,padding);
            }
            // Attach to previous widget if available
			if (prevWidgetInSet != null) {
				XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
				formAttachment.setCodAttachSide(AttachSideType.RIGHT_LITERAL);
				formAttachment.setQntOffset(padding);
				formAttachment.setAttachElement(prevWidgetInSet);
				return formAttachment;
			}
			// Attach to default position for control (which is left)
			XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
			formAttachment.setQntOffset(0);
			return formAttachment;
		} catch (RuntimeException ex) {
			ex.printStackTrace();
			throw ex;
		}
	}

	/**
	 * Creates the default right attachment
	 *
	 * @param widget
	 * @param composeData
	 * @param tab
	 * @param isLabel
	 * @param isMainWidget
	 * @param prevWidgetInSet
	 * @param nextWidgetInSet
	 * @return
	 */
	public static XMAFormAttachment createDefaultRightAttachment(XMAWidget widget, ComposeData composeData,  TabulatorPosition tab,  Boolean isLabel, Boolean isMainWidget, XMAWidget prevWidgetInSet, XMAWidget nextWidgetInSet, Boolean attachNextWidgetsLeft2) {
		try {
			//	int padding = composeData != null && composeData.getSpacing() != null ? composeData.getSpacing().getWidth() : 3;
			int padding = 3;
            FieldPart fieldPart = WidgetExtension.getFieldPartProperty(widget);
	        // ----------
			// Label
	        // ----------
			if (FieldPart.LABEL.equals(fieldPart)) {
				if (tab != null) {
	    			return createAttachmentFromTabulatorPosition(tab,0);
				}
//				// Attach to default position if it is the first widget in the set
//				if (prevWidgetInSet == null) {
//					XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
//					formAttachment.setQntOffset(130);
//					return formAttachment;
//				}

				boolean nextIsAttachedToThis = attachNextWidgetsLeft2 != null ? attachNextWidgetsLeft2==true : true ;

		        // Next Widget if available and not attatched to this (would result in opposite attachments)
				if (nextWidgetInSet != null && nextIsAttachedToThis == false) {
					XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
					formAttachment.setCodAttachSide(AttachSideType.LEFT_LITERAL);
					formAttachment.setQntOffset(-padding);
					formAttachment.setAttachElement(nextWidgetInSet);
					return formAttachment;
				}
				// Fallback is 100%
				if (nextIsAttachedToThis == false) {
					XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
					formAttachment.setQntOffset(-padding);
					formAttachment.setQntNominator(100);
					return formAttachment;
				}

				// Fallback is no attachment
				return null;
			}
	        // ----------
			// MainWidget
	        // ----------
	         if (FieldPart.CONTROL.equals(fieldPart)) {
				if (tab != null) {
	    			return createAttachmentFromTabulatorPosition(tab,0);
				}
				boolean nextIsAttachedToThis = attachNextWidgetsLeft2 != null ? attachNextWidgetsLeft2==true : true ;
				if (nextWidgetInSet != null && nextIsAttachedToThis == false) {
					XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
					formAttachment.setCodAttachSide(AttachSideType.LEFT_LITERAL);
					formAttachment.setQntOffset(-padding);
					formAttachment.setAttachElement(nextWidgetInSet);
					return formAttachment;
				}

//		        // Next Widget if available
//				if (nextWidgetInSet != null) {
//					XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
//					formAttachment.setCodAttachSide(AttachSideType.LEFT_LITERAL);
//					formAttachment.setQntOffset(-padding);
//					formAttachment.setAttachElement(nextWidgetInSet);
//					return formAttachment;
//				}
				// Fallback is 100%
				if (nextIsAttachedToThis == false) {
					XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
					formAttachment.setQntOffset(-padding);
					formAttachment.setQntNominator(100);
					return formAttachment;
				}

				// Fallback is no attachment
				return null;
			}
	        // ----------
			// NextWidget
	        //-----------
	        // Tab if available
	        if (tab != null) {
    			return createAttachmentFromTabulatorPosition(tab,0);
	        }
	        // Next Widget if available
			if (nextWidgetInSet != null) {
				XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
				formAttachment.setCodAttachSide(AttachSideType.LEFT_LITERAL);
				formAttachment.setQntOffset(-padding);
				formAttachment.setAttachElement(nextWidgetInSet);
				return formAttachment;
			}

			// Seperator
			if (widget instanceof XMASeperator) {
				return createDefaultRightAttachmentForSeperator(padding);
			}

			// Complex element
			if (isElementWhichFillCell(widget)) {
				return createDefaultRightAttachmentForFillingElement(padding);
	        }

			// Fallback is 100%
			XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
			//formAttachment.setQntOffset(-padding);
			formAttachment.setQntNominator(100);
			return formAttachment;

		} catch (RuntimeException ex) {
			ex.printStackTrace();
			throw ex;
		}
	}

	public static XMAFormAttachment createDefaultBottomAttachment(XMAWidget widget, ComposeData composeData) {
		if (isElementWhichFillCell(widget)) {
			int padding = 3;
			return createDefaultBottomAttachmentForFillingElement(padding);
		}
		return null;
	}

	private static XMAFormAttachment createDefaultRightAttachmentForSeperator(int padding) {
		XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
		formAttachment.setQntOffset(-padding);
		formAttachment.setQntNominator(100);
		return formAttachment;
	}

	private static boolean isElementWhichFillCell(XMAWidget widget) {
		if (widget instanceof XMAComposite ||
			widget instanceof XMAGroup ||
			widget instanceof XMATable ||
			widget instanceof XMATree) {
			return true;
		}
		return false;
	}

	private static XMAFormAttachment createDefaultRightAttachmentForFillingElement(int padding) {
		XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
		//formAttachment.setQntOffset(-padding);
		formAttachment.setQntNominator(100);
		return formAttachment;
	}

	private static XMAFormAttachment createDefaultTopAttachmentForFillingElement(int padding) {
		XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
		return formAttachment;
	}

	private static XMAFormAttachment createDefaultBottomAttachmentForFillingElement(int padding) {
		XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
		//formAttachment.setQntOffset(-padding);
		formAttachment.setQntNominator(100);
		return formAttachment;
	}

	private static XMAFormAttachment createAttachmentFromTabulatorPosition(TabulatorPosition tab, int padding) {
		if (tab == null) {
			throw new IllegalArgumentException("Tabulatorposition must not be null");
		}

		int offset = 0;
		int nominator = 0;
		if (tab.getReferencedPosition() != null) {
			IAttachmentPosition baseAttPos = getBaseIAttachmentPosition(tab.getReferencedPosition());
			offset = getPositionOffset(baseAttPos);
			nominator = getPositionNominator(baseAttPos);
			if (tab.getAdditionalOffset() != null) {
				offset+=tab.getAdditionalOffset().getSignedValue();
			}
		} else {
			offset = getPositionOffset(tab);
			nominator = getPositionNominator(tab);
		}
		XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
		formAttachment.setQntOffset(offset+padding);
		formAttachment.setQntNominator(nominator);
		return formAttachment;
	}

	public static boolean isButton(XMAWidget xmaWidget) {
		if (xmaWidget instanceof XMAButton && xmaWidget instanceof CheckButton==false && xmaWidget instanceof RadioButton==false) {
			return true;
		}
		return false;
	}

	/**
	 *
	 * @param setOfSimpleElements
	 * @return
	 */
	public static TabulatorPosition getLastTabulatorPosition(SetOfGuiElements setOfSimpleElements) {
		Object value = EObjectPropertiesAdapter.getProperty(setOfSimpleElements,LAST_TABULATOR_POSITION_PROPERTY_NAME);
		if (value instanceof TabulatorPosition || value == null) {
			return (TabulatorPosition)value;
		}
		throw new RuntimeException("The property '"+LAST_TABULATOR_POSITION_PROPERTY_NAME+"' is not of type 'org.openxma.xmadsl.model.TabulatorPosition'");
	}

	/**
	 *
	 * @param setOfSimpleElements
	 * @param leftTabulatorPosition
	 * @param rightTabulatorPosition
	 */
	public static void setLastTabulatorPosition(SetOfGuiElements setOfSimpleElements, TabulatorPosition tabulatorPosition) {
		if (tabulatorPosition != null) {
			EObjectPropertiesAdapter.setProperty(setOfSimpleElements,LAST_TABULATOR_POSITION_PROPERTY_NAME,tabulatorPosition);
		} else {
			EObjectPropertiesAdapter.removeProperty(setOfSimpleElements,LAST_TABULATOR_POSITION_PROPERTY_NAME);
		}
	}

	/**
	 * Determine the FieldPart 'CONTROL' for the given GuiElement
	 *
	 * @param groupWidgetList
	 * @param guiElement
	 * @return
	 */
	public static XMAWidget getControlFieldPartForGuiElement(List<XMAWidget> groupWidgetList, GuiElement guiElement) {
		if (guiElement == null) {
			throw new RuntimeException("GuiElement has to be defined");
		}
		for (XMAWidget widget : groupWidgetList) {
			GuiElement tGE = WidgetExtension.getGuiElementProperty(widget);
			if (tGE == guiElement) {
				FieldPart fieldPart = WidgetExtension.getFieldPartProperty(widget);
				if (FieldPart.CONTROL.equals(fieldPart)) {
                    return widget;
                }
			}
		}
		return null;
	}

	/**
	 * Create the top-Attachment within a SetOfWidget.
	 *
	 * @param topAttachment
	 * @param parentComposite
	 * @param centeredWidget
	 * @param controlFieldPart
	 * @param xmaWidget
	 * @return
	 */
	public static XMAFormAttachment createTopAttachment(AttachmentSpecification topAttachment, XMAComposite parentComposite, GuiElement guiElement,  GuiElement prevGuiElement, GuiElement nextGuiElement, XMAWidget prevWidget, XMAWidget centeredWidget, XMAWidget controlFieldPart, XMAWidget xmaWidget) {

		boolean thisIsControlFieldPart = controlFieldPart == xmaWidget;

		// Only use top attchment of GuiElement in case the controlFieldPart is available
		if (topAttachment != null && (thisIsControlFieldPart || controlFieldPart == null)) {
			return createXMAAttachment(topAttachment,parentComposite,prevGuiElement,nextGuiElement,AttachmentOwnPosition.TOP);
		}

		// Default for a Complex element which is the only widget of a cell
		if (isElementWhichFillCell(xmaWidget) && prevGuiElement == null && nextGuiElement == null) {
			return createDefaultTopAttachmentForFillingElement(3);
        }

		List<Style> styleList = PresentationModelExtension.getStyles(guiElement);
		XMAFormAttachment layoutStrategyAttachment = GeneratorProperties.getLayoutStrategy().getDefaultTopAttachmentInSet(xmaWidget,prevWidget,centeredWidget,controlFieldPart,styleList);
		if (layoutStrategyAttachment != null) {
			return layoutStrategyAttachment;
		} else {
			// Only attach to controlFieldPart if it is not itself
			XMAWidget widgetForCenterAttachment = topAttachment != null && controlFieldPart != null && thisIsControlFieldPart == false
				? controlFieldPart
				: centeredWidget;
			XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
			formAttachment.setCodAttachSide(AttachSideType.CENTER_LITERAL);
			formAttachment.setAttachElement(widgetForCenterAttachment);
			return formAttachment;
		}
	}

	public static void createMargin(XMAWidget xmaWidget, ComposeData composeData) {
		if (xmaWidget instanceof XMAComposite && composeData != null && composeData.getMargin() != null) {
			int margin = composeData.getMargin().getWidth();
			XMAComposite xmaComposite = (XMAComposite)xmaWidget;
			xmaComposite.setQntMarginHeight(margin);
			xmaComposite.setQntMarginWidth(margin);
		}
	}

}
