package org.openxma.dsl.generator.helper;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.openxma.dsl.pom.model.Component;
import org.openxma.dsl.pom.model.Page;

import at.spardat.xma.guidesign.XMAComposite;
import at.spardat.xma.guidesign.XMADialogPage;
import at.spardat.xma.guidesign.XMAPage;
import at.spardat.xma.guidesign.XMATable;

public class PageExtension {
	
	public static List<XMATable> getAllContainedTablesWithinPage(EObject xmaPage) {
		return getAllContainedTablesWithinPage(xmaPage,null);
	}
	public static List<XMATable> getAllContainedTablesWithinPage(EObject xmaPage, List<XMATable> parentTableList) {
		List<EObject> content = xmaPage.eContents();		
		List<XMATable> tableList = null;
		// Reuse given list or create a new one 
		if (parentTableList != null) {
			tableList = parentTableList;
		} else {
			tableList = new ArrayList<XMATable>();
		}
		// Fill each table into list
		for (EObject eObj : content) {
			if (eObj instanceof XMATable) {
				tableList.add((XMATable)eObj);				
			} else if (eObj.eContents() != null && eObj instanceof XMAPage == false) {
				getAllContainedTablesWithinPage(eObj,tableList);
			}
		}
		return tableList;		
	}
	
	/**
	 * Create a List with all Pages contained by a Component ordered hierarchical.
	 * 
	 * @param component
	 * @return
	 */
	public static List<Page> getPagesOfComponentHierarchical(Component component) {
		ArrayList<Page> pageList = new ArrayList<Page>();
		collectPagesHierachical(component,pageList);
		return pageList;
	}
	
	/**
	 * Collects recursivly all Pages owned by obj and store them into pageList.  
	 * 
	 * @param obj
	 * @param pageList
	 */
	private static void collectPagesHierachical(EObject obj, List<Page> pageList) {
		if (obj != null && obj.eContents() != null) {
			for (EObject childObj : obj.eContents()) {
				if (childObj instanceof Page) {
					pageList.add((Page)childObj);
				}
				collectPagesHierachical(childObj,pageList);					
			}
		}	
	}
	
	/**
	 * Collects all Composites which are contained by a Page and which have to
	 * be extended.
	 * 
	 * @param xmaPage
	 * @return
	 */
	public static List<XMAComposite> getCompositesToBeExtended(XMAPage xmaPage) {
		
		List<XMAComposite> compositesToBeExtended = new ArrayList<XMAComposite>();
		boolean pageIsGenerated = (xmaPage instanceof XMADialogPage) 
			? ((XMADialogPage)xmaPage).isYnGenerated() 
			: false;
			
		TreeIterator<EObject> iterator = xmaPage.eAllContents();		
		while (iterator.hasNext()) {
			EObject obj = iterator.next();
			if (obj instanceof XMAPage) {
				iterator.prune();
				continue;
			}				
			if (obj instanceof XMAComposite && (pageIsGenerated || ((XMAComposite)obj).isYnSimpleLayout())) {
				compositesToBeExtended.add((XMAComposite)obj);
			}
		}			
		
		return compositesToBeExtended;
	}

}
