package org.openxma.dsl.generator.impl.substrategy;

import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.openxma.dsl.common.Strings;
import org.openxma.dsl.common.extensions.EObjectPropertiesAdapter;
import org.openxma.dsl.core.model.TrueLiteral;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.Entity;
import org.openxma.dsl.dom.model.FeatureReference;
import org.openxma.dsl.generator.helper.EventExtension;
import org.openxma.dsl.generator.helper.PresentationModelExtension;
import org.openxma.dsl.generator.helper.WidgetExtension;
import org.openxma.dsl.generator.impl.DomPageLogicAttachmentStrategy;
import org.openxma.dsl.pom.PomPackage;
import org.openxma.dsl.pom.model.GuiElement;
import org.openxma.dsl.pom.model.StylePropertyTableCustomizer;
import org.openxma.dsl.pom.model.StyleSpecificationProperty;
import org.openxma.dsl.pom.model.Table;
import org.openxma.dsl.pom.model.TableColumn;
import org.openxma.dsl.pom.scoping.PomElementCollector;

import at.spardat.xma.guidesign.XMAPage;
import at.spardat.xma.guidesign.XMAPagingControl;
import at.spardat.xma.guidesign.XMATable;
import at.spardat.xma.guidesign.XMATableColumn;

public class TableCustomizerStrategy {

    public List<XMATable> getXMATables(XMAPage page) {
        return PomElementCollector.getElementsOfXMAPage(page,XMATable.class);
    }

    public void genCreateWidgetsCodeClient(DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy, XMAPage page, StringBuilder buf) {
        List<XMATable> tablesOfPage = getXMATables(page);
        for (XMATable xmaTable : tablesOfPage) {
            GuiElement guiElement = WidgetExtension.getGuiElementProperty(xmaTable);
            if (!(guiElement instanceof Table)) continue;
            Table table = (Table)guiElement;
            String tableCustomizerInterface = null;
            String tableCustomizerName = xmaTable.getNamInstance()+"_customizer";
            if (table.getPageable() instanceof TrueLiteral) {
                Map<Integer,StyleSpecificationProperty> styleSpecificationProperties = PresentationModelExtension.getCombinedStyleProperty(table);
                if (styleSpecificationProperties != null) {
                    StylePropertyTableCustomizer customizerProp = (StylePropertyTableCustomizer)styleSpecificationProperties.get(PomPackage.STYLE_PROPERTY_TABLE_CUSTOMIZER);
                    if (customizerProp != null) {
                        tableCustomizerInterface = customizerProp.getTableCustomizer().getInstanceType();
                        int lastDotIndex = tableCustomizerInterface.lastIndexOf('.');
                        if (lastDotIndex >= 0) {
                            tableCustomizerInterface = tableCustomizerInterface.substring(lastDotIndex+1);
                        }
                    }
                }
                if (tableCustomizerInterface == null) {
                    continue;
                }
                XMAPagingControl pc = EventExtension.getPagingControlForXMATable(xmaTable);
                String xmaTableName = xmaTable.getNamInstance();
                buf.append("\n        /* Table customizer for "+xmaTableName+" */");
                buf.append("\n        "+tableCustomizerInterface+" "+tableCustomizerName+" = ("+tableCustomizerInterface+") getTypedComponent().getSession().getComponent("+tableCustomizerInterface+".COMPONENT_NAME);");
                buf.append("\n        "+tableCustomizerName+".setTableLayoutManager("+xmaTable.getNamInstance()+"WLM);");
                buf.append("\n        "+tableCustomizerName+".setTableWMClient("+xmaTable.getNamInstance()+");");
                buf.append("\n        "+tableCustomizerName+".setFQTableName(\""+page.getComponent().getNamPackage()+"."+page.getNamClass() +"."+xmaTableName+"\");");
                buf.append("\n        Map<String, ColumnMetaData> "+xmaTableName+"_columnMap = new HashMap<String, ColumnMetaData>();");
                for (Object obj: xmaTable.getColumn()) {
                    XMATableColumn xmaColumn = (XMATableColumn)obj;
                    TableColumn xmadslColumn = (TableColumn)EObjectPropertiesAdapter.getProperty(xmaColumn,"xmadslColumn");
                    if (xmadslColumn == null || xmadslColumn.getFeature() == null) {
                        continue; // Do not add pure UI columns (without linked attributes)
                    }
                    EObject attributeHolder = xmadslColumn.getFeature().getAttributeHolder();
                    Attribute attribute = null;
                    if (attributeHolder instanceof FeatureReference) {
                        FeatureReference featureReference = (FeatureReference)attributeHolder;
                        if (featureReference.getSource() instanceof Entity) {
                            attribute = featureReference.getAttribute();
                        }
                    } else if (attributeHolder instanceof Attribute) {
                        if (attributeHolder.eContainer() instanceof Entity) {
                            attribute = (Attribute)attributeHolder;
                        }
                    }

                    if (attribute != null) {
                        buf.append("\n        "+xmaTableName+"_columnMap.put(\""+attribute.getName()+"\",new ColumnMetaData(\""+attribute.getName().toUpperCase()+"\", messages.getString(\""+xmaColumn.getNamRscKeyLabel()+"\"), "+xmaColumn.getNamInstance().toUpperCase()+"));");
                    }
                }
                buf.append("\n        "+tableCustomizerName+".setColumnMap("+xmaTableName+"_columnMap);");
                buf.append("\n        "+tableCustomizerName+".setAppCode(getAppCode());");
                buf.append("\n        "+tableCustomizerName+".setPagingWMClient("+xmaTableName+"_pagingControl);");
                buf.append("\n        "+tableCustomizerName+".init();");
                buf.append("\n        "+xmaTable.getNamInstance()+"_customizerComponentId.set((short)"+tableCustomizerName+".getId());");
                buf.append("\n        "+pc.getNamInstance()+".setCustomizer("+tableCustomizerName+");");
                buf.append("\n");
            }
        }
    }

    public void genAdditionalImportsClient(DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy, XMAPage xmaPage, LinkedHashSet<String> imports) {
        List<XMATable> tablesOfPage = getXMATables(xmaPage);
        for (XMATable xmaTable : tablesOfPage) {
            GuiElement guiElement = WidgetExtension.getGuiElementProperty(xmaTable);
            if (!(guiElement instanceof Table)) continue;
            Table table = (Table)guiElement;
            if (table.getPageable() instanceof TrueLiteral) {
                Map<Integer,StyleSpecificationProperty> styleSpecificationProperties = PresentationModelExtension.getCombinedStyleProperty(table);
                if (styleSpecificationProperties != null) {
                    StylePropertyTableCustomizer customizerProp = (StylePropertyTableCustomizer)styleSpecificationProperties.get(PomPackage.STYLE_PROPERTY_TABLE_CUSTOMIZER);
                    if (customizerProp != null) {
                        String customizerType = customizerProp.getTableCustomizer().getInstanceType();
                        imports.add(customizerType);
                        imports.add("java.util.Map");
                        imports.add("java.util.HashMap");
                        imports.add("org.openxma.addons.ui.table.customizer.common.ColumnMetaData");
                    }
                }
            }
        }
    }

    public void genAdditionalImportsServer(DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy, XMAPage xmaPage, LinkedHashSet<String> imports) {
        List<XMATable> tablesOfPage = getXMATables(xmaPage);
        for (XMATable xmaTable : tablesOfPage) {
            GuiElement guiElement = WidgetExtension.getGuiElementProperty(xmaTable);
            if (!(guiElement instanceof Table)) continue;
            Table table = (Table)guiElement;
            if (table.getPageable() instanceof TrueLiteral) {
                Map<Integer,StyleSpecificationProperty> styleSpecificationProperties = PresentationModelExtension.getCombinedStyleProperty(table);
                if (styleSpecificationProperties != null) {
                    StylePropertyTableCustomizer customizerProp = (StylePropertyTableCustomizer)styleSpecificationProperties.get(PomPackage.STYLE_PROPERTY_TABLE_CUSTOMIZER);
                    if (customizerProp != null) {
                        imports.add("org.openxma.dsl.platform.query.QueryObject");
                        imports.add("at.spardat.xma.session.XMASessionServer");
                        imports.add(getTableCustomizerCompName(customizerProp.getTableCustomizer().getInstanceType(),true));
                    }
                }
            }
        }
    }

    public void genAdditionalMemberVariablesClient(DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy, XMAPage xmaPage, StringBuilder stringBuilder) {
    }

    public void genAdditionalMemberVariablesServer(
            DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy,
            XMAPage xmaPage, StringBuilder stringBuilder) {

    }

    protected String getTableCustomizerCompName(String tableCustomizerInterfaceName, boolean fullQualified) {
        if (tableCustomizerInterfaceName == null) {
            throw new IllegalArgumentException("The value of tableCustomizerInterfaceName must not be null");
        }
        String tableCustomizername;
        int lastDotIndex = tableCustomizerInterfaceName.lastIndexOf('.');
        if (lastDotIndex >= 0) {
            tableCustomizername = tableCustomizerInterfaceName.substring(lastDotIndex+1);
        } else {
            tableCustomizername = tableCustomizerInterfaceName;
        }
        tableCustomizername+="Server";
        if (fullQualified && lastDotIndex >= 0) {
            String clientPackage = tableCustomizerInterfaceName.substring(0,lastDotIndex);
            if (clientPackage.endsWith(".client")) {
                int clientDotIndex = clientPackage.lastIndexOf(".client");
                String serverPackage = clientPackage.substring(0,clientDotIndex)+".server.";
                tableCustomizername = serverPackage+tableCustomizername;
            } else {
                tableCustomizername = clientPackage+"."+tableCustomizername;
            }
        }
        return tableCustomizername;
    }

    public void genGetQueryObject(DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy, XMAPage page, StringBuilder buffer) {

         List<XMATable> tablesOfPage = getXMATables(page);
            for (XMATable xmaTable : tablesOfPage) {
                GuiElement guiElement = WidgetExtension.getGuiElementProperty(xmaTable);
                if (!(guiElement instanceof Table)) continue;
                Table table = (Table)guiElement;
                String tableCustomizername = null;
                if (table.getPageable() instanceof TrueLiteral) {
                    Map<Integer,StyleSpecificationProperty> styleSpecificationProperties = PresentationModelExtension.getCombinedStyleProperty(table);
                    if (styleSpecificationProperties != null) {
                        StylePropertyTableCustomizer customizerProp = (StylePropertyTableCustomizer)styleSpecificationProperties.get(PomPackage.STYLE_PROPERTY_TABLE_CUSTOMIZER);
                        if (customizerProp != null) {
                            tableCustomizername = getTableCustomizerCompName(customizerProp.getTableCustomizer().getInstanceType(),false);
                        }
                    }
                    if (tableCustomizername == null) {
                        continue;
                    }
                    XMAPagingControl pc = EventExtension.getPagingControlForXMATable(xmaTable);
                    buffer.append(
                        "\n    /**\n"+
                        "     * Gets the QueryObject for the table <tt>"+xmaTable.getNamInstance()+"</tt>\n"+
                        "     */\n"+
                        "     public QueryObject get"+Strings.capitalize(xmaTable.getNamInstance())+"QueryObject() {\n"+
                        "         int offset = "+pc.getNamInstance()+".getOffset();\n"+
                        "         short pageSize = "+pc.getNamInstance()+".getPageSize();\n"+

                        "         short sortingColumn = "+pc.getNamInstance()+".getSortingColumn();\n"+
                        "         boolean isAscending = "+pc.getNamInstance()+".getAscending();\n"+
                        "         XMASessionServer xmaSessionServer = (XMASessionServer)XMASessionServer.getXMASession();\n"+
                        "         "+tableCustomizername+" tableCustomizer = ("+tableCustomizername+")xmaSessionServer.getComponent(Short.valueOf("+xmaTable.getNamInstance()+"_customizerComponentId.getEncodedValue()));\n"+
                        "         QueryObject filterQueryForTable = null;\n"+
                        "         if (sortingColumn >-1) {\n"+
                        "             filterQueryForTable = tableCustomizer.getQueryObject(sortingColumn,isAscending);\n"+
                        "         } else {\n"+
                        "             filterQueryForTable = tableCustomizer.getQueryObject();\n"+
                        "         }\n"+
                        "         filterQueryForTable.setFirstResult(offset);\n"+
                        "         filterQueryForTable.setMaxResults(pageSize);\n"+
                        "         return filterQueryForTable;\n"+
                        "     }\n");
                }
            }

    }
}
